//|------------------------------------------------------------------+
//|                                               Unity-Forecast.mq5 |
//|                                    Copyright (c) 2018, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//|                           https://www.mql5.com/ru/articles/5473/ |
//|------------------------------------------------------------------+
#property copyright "2018 © Marketeer"
#property link      "https://www.mql5.com/en/users/marketeer"
#property version   "1.0"
#property description "Multi-asset cluster indicator taking all currencies as a sum of squares forming market unity (1.0)"
#property description "Includes SOM forecast"

// specify available filename to load NetFileName in the tester (without on-the-fly training)
// #property tester_file "offline-training-result.som"

#define BUF_NUM 16

// #define SOM_VERBOSE
#define SOM_OUTLIERS_SOFT

#property indicator_separate_window
#property indicator_buffers BUF_NUM
#property indicator_plots BUF_NUM

#property indicator_color1 Green
#property indicator_color2 DarkBlue
#property indicator_color3 Red
#property indicator_color4 Gray
#property indicator_color5 Peru
#property indicator_color6 Gold
#property indicator_color7 Purple
#property indicator_color8 Teal

#property indicator_color9 LightGreen
#property indicator_color10 LightBlue
#property indicator_color11 Orange
#property indicator_color12 LightGray
#property indicator_color13 Brown
#property indicator_color14 Yellow
#property indicator_color15 Magenta

#property indicator_color16 Aqua


input string _1 = ""; // Cluster Indicator Settings
input string Instruments = "EURUSD,GBPUSD,USDCHF,USDJPY,AUDUSD,USDCAD,NZDUSD,XAUUSD"; // ·    Instruments
input int BarLimit = 352; // ·    BarLimit (training bars + BarLookback + 1 extra)
input ENUM_DRAW_TYPE Draw = DRAW_LINE; // ·    Draw
input ENUM_APPLIED_PRICE PriceType = PRICE_CLOSE; // ·    PriceType
const bool AbsoluteValues = false;

input string _2 = ""; // SOM Network and Data Settings
input string NetFileName = ""; // ·    NetFileName (*.som)
input int CellsX = 10; // ·    CellsX
input int CellsY = 10; // ·    CellsY
input bool HexagonalCell = true; // ·    HexagonalCell
input bool UseNormalization = true; // ·    UseNormalization
input int EpochNumber = 100; // ·    EpochNumber
input int RetrainingBars = 20; // ·    RetrainingBars
input int BarLookback = 1; // ·    BarLookback
input bool ForecastMode = true; // ·    ForecastMode

input string _3 = ""; // Options
input bool ShowProgress = true; // ·    ShowProgress
input int RandomSeed = 0; // ·    RandomSeed
input bool SaveTrainedNetworks = false; // ·    SaveTrainedNetworks

int ShiftedBuffer = -1;

bool initDone = false;

#include <CSOM/IndArray.mqh>
#include <CSOM/HashMapTemplate.mqh>
#include <CSOM/CSOM.mqh>

IndicatorArray buffers(BUF_NUM);
IndicatorArrayGetter getter(buffers);

string Symbols[];
int Direction[];
double Contracts[];
int SymbolCount;

HashMapTemplate<string,int> workCurrencies;
int baseIndex = 0;
string BaseCurrency;
int LastBarCount;

bool initSymbols()
{
  SymbolCount = StringSplit(Instruments, ',', Symbols);
  if(SymbolCount >= BUF_NUM)
  {
    SymbolCount = BUF_NUM - 1;
    ArrayResize(Symbols, SymbolCount);
  }
  else if(SymbolCount == 0)
  {
    SymbolCount = 1;
    ArrayResize(Symbols, SymbolCount);
    Symbols[0] = Symbol();
  }
  
  ArrayResize(Direction, SymbolCount);
  ArrayInitialize(Direction, 0);
  ArrayResize(Contracts, SymbolCount);
  ArrayInitialize(Contracts, 1);
  
  string base = NULL;
  
  for(int i = 0; i < SymbolCount; i++)
  {
    if(!SymbolSelect(Symbols[i], true))
    {
      Print("Can't select ", Symbols[i]);
      return false;
    }
    
    string first, second;
    first = SymbolInfoString(Symbols[i], SYMBOL_CURRENCY_BASE);
    second = SymbolInfoString(Symbols[i], SYMBOL_CURRENCY_PROFIT);
    
    if(first != second)
    {
      workCurrencies.inc(first, 1);
      workCurrencies.inc(second, 1);
    }
    else
    {
      workCurrencies.inc(Symbols[i], 1);
    }
  }
  
  if(workCurrencies.getSize() >= BUF_NUM)
  {
    Print("Too many symbols, max ", (BUF_NUM - 1));
    return false;
  }
  
  ShiftedBuffer = workCurrencies.getSize() - 1;
  
  for(int i = 0; i < workCurrencies.getSize(); i++)
  {
    if(workCurrencies[i] > 1)
    {
      if(base == NULL)
      {
        base = workCurrencies.getKey(i);
        baseIndex = i;
      }
      else
      {
        Print("Collision: multiple base symbols: ", base, "[", workCurrencies[base], "] ", workCurrencies.getKey(i), "[", workCurrencies[i], "]");
        return false;
      }
    }
  }
  
  BaseCurrency = base;
  
  for(int i = 0; i < SymbolCount; i++)
  {
    if(SymbolInfoString(Symbols[i], SYMBOL_CURRENCY_PROFIT) == base) Direction[i] = +1;
    else if(SymbolInfoString(Symbols[i], SYMBOL_CURRENCY_BASE) == base) Direction[i] = -1;
    else
    {
      Print("Ambiguous symbol direction ", Symbols[i], ", defaults used");
      Direction[i] = +1;
    }
    
    Contracts[i] = SymbolInfoDouble(Symbols[i], SYMBOL_TRADE_CONTRACT_SIZE);
  }
  
  // ArrayPrint(Symbols);
  // ArrayPrint(Contracts);
  
  return true;
}

int OnInit()
{
  ShiftedBuffer = -1;
  initDone = false;
  
  if(!initSymbols()) return INIT_PARAMETERS_INCORRECT;

  MathSrand(RandomSeed);
  
  string base = SymbolInfoString(_Symbol, SYMBOL_CURRENCY_BASE);
  string profit = SymbolInfoString(_Symbol, SYMBOL_CURRENCY_PROFIT);
  
  int replaceIndex = -1;
  for(int i = 0; i <= SymbolCount; i++)
  {
    string name;
    if(i == 0)
    {
      name = BaseCurrency;
      if(name != workCurrencies.getKey(i))
      {
        replaceIndex = i;
      }
    }
    else
    {
      if(BaseCurrency == workCurrencies.getKey(i) && replaceIndex > -1)
      {
        name = workCurrencies.getKey(replaceIndex);
      }
      else
      {
        name = workCurrencies.getKey(i);
      }
    }
    
    PlotIndexSetString(i, PLOT_LABEL, name);
    int width = (name == base || name == profit) ? 2 : 1;
    PlotIndexSetInteger(i, PLOT_DRAW_TYPE, Draw);
    PlotIndexSetInteger(i, PLOT_LINE_WIDTH, width);
    PlotIndexSetInteger(i, PLOT_SHOW_DATA, true);
  }

  PlotIndexSetString(SymbolCount + 1, PLOT_LABEL, "Forecast");
  PlotIndexSetInteger(SymbolCount + 1, PLOT_DRAW_TYPE, Draw);
  PlotIndexSetInteger(SymbolCount + 1, PLOT_LINE_WIDTH, 5);
  PlotIndexSetInteger(SymbolCount + 1, PLOT_LINE_COLOR, clrAqua);

  PlotIndexSetInteger(SymbolCount + 1, PLOT_SHOW_DATA, true);
  // PlotIndexSetInteger(SymbolCount + 1, PLOT_SHIFT, _0_); // show Close[0] at opening of 0-bar: forecast!
  
  for(int i = SymbolCount + 2; i < BUF_NUM; i++)
  {
    PlotIndexSetInteger(i, PLOT_SHOW_DATA, false);
  }
  
  if(!AbsoluteValues)
  {
    IndicatorSetInteger(INDICATOR_LEVELS, 1);
    IndicatorSetDouble(INDICATOR_LEVELVALUE, 0, 0.0);
  }
  
  IndicatorSetString(INDICATOR_SHORTNAME, "Unity Forecast [" + (string)workCurrencies.getSize() + "] " + EnumToString(PriceType));
  IndicatorSetInteger(INDICATOR_DIGITS, 5);
  
  LastBarCount = 0;
  initDone = true;
  
  if(!MQLInfoInteger(MQL_TESTER))
  {
    EventSetTimer(1);
  }

  Print("Unity-Forecast ", __FUNCTION__, " ", Bars(_Symbol, PERIOD_CURRENT));

  return INIT_SUCCEEDED;
}


#define CopyXYZ(N, S, B, A) if(Copy##N(S, PERIOD_CURRENT, B, 1, A) != 1) {Print("Failed Copy", #N, " ", S, " ", B, " ", GetLastError()); return false;}
#define CopyXYZS(N, B, A) if(Copy##N(_Symbol, PERIOD_CURRENT, B, 1, A) != 1) {Print("Failed Copy", #N, " ", _Symbol, " ", B, " ", GetLastError()); return false;}
#define CopyXYZN(N, B, C, A) if(Copy##N(_Symbol, PERIOD_CURRENT, B, C, A) != C) {Print("Failed Copy", #N, " ", _Symbol, " ", B, " ", GetLastError()); return false;}
#define CopyXYZSN(N, S, B, C, A) if(Copy##N(S, PERIOD_CURRENT, B, C, A) != C) {Print("Failed Copy", #N, " ", S, " ", B, " ", GetLastError()); return false;}

#define TOSTRING(A) #A + " = " + (string)(A) + " "

bool calculate(const int bar)
{
  datetime time[1];
  CopyXYZN(Time, bar, 1, time);
  
  double w[2] = {0};
  double v[][2];
  ArrayResize(v, SymbolCount);
  
  for(int j = 0; j < SymbolCount; j++)
  {
    if(PriceType == PRICE_OPEN)
    {
      CopyXYZSN(Open, Symbols[j], time[0], 2, w);
    }
    else
    if(PriceType == PRICE_CLOSE)
    {
      CopyXYZSN(Close, Symbols[j], time[0], 2, w);
    }

    if(Direction[j] == -1)
    {
      w[0] = 1.0 / w[0];
      w[1] = 1.0 / w[1];
    }

    w[1] *= Contracts[j];
    w[0] *= Contracts[j];
    
    v[j][0] = w[1];
    v[j][1] = w[0];
  }
  
  double sum[2] = {1.0, 1.0};
  
  for(int j = 0; j < SymbolCount; j++)
  {
    sum[0] += v[j][0] * v[j][0];
    sum[1] += v[j][1] * v[j][1];
  }
  
  double base2_0 = 1.0 / sum[0];
  double base2_1 = 1.0 / sum[1];
  
  double base_0 = MathSqrt(base2_0);
  double base_1 = MathSqrt(base2_1);
  
  if(AbsoluteValues)
  {
    buffers[0][bar] = base_0 * Contracts[0];
  }
  else
  {
    buffers[0][bar] = base_0 / base_1 - 1.0;
  }
  
  for(int j = 1; j <= SymbolCount; j++)
  {
    if(AbsoluteValues)
    {
      buffers[j][bar] = base_0 * v[j - 1][0];
    }
    else
    {
      buffers[j][bar] = base_0 * v[j - 1][0] / (base_1 * v[j - 1][1]) - 1.0;
    }
  }
  
  return true;
}

void OnTimer()
{
  const double price[] = {};
  
  if(LastBarCount != Bars(_Symbol, PERIOD_CURRENT))
  {
    int n = OnCalculate(Bars(_Symbol, PERIOD_CURRENT), 0, 0, price);
    if(n > 0)
    {
      EventKillTimer();
      ChartSetSymbolPeriod(0, _Symbol, PERIOD_CURRENT);
      ChartRedraw();
    }
  }
  else
  {
    EventKillTimer();
  }
}

int OnCalculate(const int rates_total,
                const int prev_calculated,
                const int begin,
                const double& price[])
{
  if(!initDone) return 0;
  
  if(LastBarCount == rates_total && prev_calculated == rates_total && PriceType != PRICE_CLOSE)
  {
    return prev_calculated;
  }
  
  if(prev_calculated == 0)
  {
    for(int i = 0; i < rates_total; i++)
    {
      for(int j = 0; j < BUF_NUM; j++)
      {
        buffers[j][i] = EMPTY_VALUE;
      }
    }
  }
  
  int limit = MathMin(rates_total - prev_calculated + 1, BarLimit);

  for(int i = 0; i < limit; i++)
  {
    if(!calculate(i))
    {
      return 0; // will retry on next tick
    }
  }
  
  if(LastBarCount != rates_total || prev_calculated != rates_total)
  {
    static int prev_training = 0;
    if(prev_training == 0 || prev_calculated - prev_training > RetrainingBars)
    {
      if(NetFileName != "")
      {
        if(!KohonenMap.Load(NetFileName))
        {
          Print("Map loading failed: ", NetFileName);
          initDone = false;
          return 0;
        }
      }
      else
      {
        TrainSOM(BarLimit);
      }
      prev_training = prev_calculated > 0 ? prev_calculated : rates_total;
    }
    ForecastBySOM(prev_calculated == 0);
  }
  
  LastBarCount = rates_total;
  
  return LastBarCount;
}

double GetBuffer(int index, int bar)
{
  return getter[index][bar];
}

CSOM KohonenMap;

bool ForecastBySOM(const bool anew = false)
{
  // Print(__FUNCTION__);

  double _vector[], forecast[];
  
  int n = workCurrencies.getSize();
  ArrayResize(_vector, BarLookback * n + 1);
  
  for(int k = BarLookback - 1; k >= 0; k--)
  {
    for(int j = 0; j < n; j++)
    {
      _vector[j + (BarLookback - 1 - k) * n] = GetBuffer(j, k + 1); // 1-st bar is the latest completed
    }
  }
  _vector[n * BarLookback] = 0;
  
  KohonenMap.SetFeatureMask(KohonenMap.GetFeatureCount() - 1, 0);
  KohonenMap.GetBestMatchingFeatures(_vector, forecast);
  
  static int total = 0;
  static int correct = 0;
  static double error = 0;
  static double variance = 0;
  
  buffers[n][0] = forecast[BarLookback * n];
  if(anew)
  {
    buffers[n][1] = GetBuffer(n - 1, 1);

    // prediction performance
    total = 0;
    correct = 0;
    error = 0;
    variance = 0;
  }
  else
  {
    total++;

    double predicted = GetBuffer(n, 1);
    double actual = GetBuffer(n - 1, 1);
    // Print(TimeCurrent(), " ", predicted, " ", actual);
    if(predicted * actual > 0)
    {
      correct++;
    }
    error += (predicted - actual) * (predicted - actual);
    variance += actual * actual;
    Comment("Correct ", correct, " from ", total, " ", DoubleToString(correct * 100.0 / total, 2), "%, ", (float)(error/variance));
  }
  
  return true;
}

bool LoadPatterns(const int limit)
{
  Print(__FUNCTION__);
  
  string titles[];
  int n = workCurrencies.getSize();
  ArrayResize(titles, BarLookback * n + 1);
  for(int k = BarLookback - 1; k >= 0; k--)
  {
    for(int i = 0; i < n; i++)
    {
      titles[i + (BarLookback - 1 - k) * n] = workCurrencies.getKey(i) + (string)(k + 1);
    }
  }
  titles[n * BarLookback] = "FORECAST";
  
  KohonenMap.AssignFeatureTitles(titles);

  double x[];
  ArrayResize(x, BarLookback * n + 1);
  
  for(int i = limit - BarLookback; i >= 1; i--) // 0-th bar is incomplete
  {
    datetime time[1];
    CopyXYZN(Time, i, 1, time);
    for(int k = BarLookback - 1; k >= 0; k--)
    {
      for(int j = 0; j < n; j++)
      {
        x[j + (BarLookback - 1 - k) * n] = GetBuffer(j, i + k);
      }
    }
    x[n * BarLookback] = GetBuffer(n - 1, i - 1); // last listed currency's next bar
    KohonenMap.AddPattern(x, (string)time[0]);
  }
  
  Print("Data count: ", KohonenMap.GetDataCount());

  return true;
}

bool TrainSOM(const int limit)
{
  Print(__FUNCTION__);
  
  KohonenMap.Reset();

  LoadPatterns(limit);
  
  KohonenMap.Init(CellsX, CellsY, HexagonalCell);
  if(ForecastMode)
  {
    KohonenMap.SetFeatureMask(KohonenMap.GetFeatureCount() - 1, 0);
  }

  KohonenMap.Train(EpochNumber, UseNormalization, ShowProgress);

  if(SaveTrainedNetworks)  
  {
    KohonenMap.Save("online-" + _Symbol + CSOM::timestamp() + ".som");
  }
  
  return true;
}
